"""Generic utility methods not part of main API."""

from __future__ import annotations

import builtins
from collections.abc import Hashable, Iterable, Mapping, Sequence
from decimal import Decimal
from functools import wraps
import inspect
import operator
import warnings

import pydash as pyd


#: Singleton object that differentiates between an explicit ``None`` value and an unset value.
#: As a class so it has its own type
class Unset: ...


UNSET = Unset()

#: Tuple of number types.
NUMBER_TYPES = (int, float, Decimal)

#: Dictionary of builtins with keys as the builtin function and values as the string name.
BUILTINS = {value: key for key, value in builtins.__dict__.items() if isinstance(value, Hashable)}

#: Object keys that are restricted from access via path access.
RESTRICTED_KEYS = ("__globals__", "__builtins__")


def callit(iteratee, *args, **kwargs):
    """Inspect argspec of `iteratee` function and only pass the supported arguments when calling
    it."""
    maxargs = len(args)
    argcount = kwargs["argcount"] if "argcount" in kwargs else getargcount(iteratee, maxargs)
    argstop = min([maxargs, argcount])

    return iteratee(*args[:argstop])


def getargcount(iteratee, maxargs):
    """Return argument count of iteratee function."""
    if hasattr(iteratee, "_argcount"):
        # Optimization feature where argcount of iteratee is known and properly
        # set by initiator.
        # It should always be right, but it can be `None` for the function wrappers
        # in `pydash.function` as the wrapped functions are out of our control and
        # can support an unknown number of arguments.
        argcount = iteratee._argcount
        return argcount if argcount is not None else maxargs

    if isinstance(iteratee, type) or pyd.is_builtin(iteratee):
        # Only pass single argument to type iteratees or builtins.
        argcount = 1
    else:
        argcount = 1

        try:
            argcount = _getargcount(iteratee, maxargs)
        except TypeError:  # pragma: no cover
            pass

    return argcount


def _getargcount(iteratee, maxargs):
    argcount = None

    try:
        # PY2: inspect.signature was added in Python 3.
        # Try to use inspect.signature when possible since it works better for our purpose of
        # getting the iteratee argcount since it takes into account the "self" argument in callable
        # classes.
        sig = inspect.signature(iteratee)
    except (TypeError, ValueError, AttributeError):  # pragma: no cover
        pass
    else:
        if not any(
            param.kind == inspect.Parameter.VAR_POSITIONAL for param in sig.parameters.values()
        ):
            argcount = len(sig.parameters)

    if argcount is None:
        # Signatures were added these operator methods in Python 3.12.3 and 3.11.9 but their
        # instance objects are incorrectly reported as accepting varargs when they only accept a
        # single argument.
        if isinstance(iteratee, (operator.itemgetter, operator.attrgetter, operator.methodcaller)):
            argcount = 1
        else:
            argspec = inspect.getfullargspec(iteratee)
            if argspec and not argspec.varargs:  # pragma: no cover
                # Use inspected arg count.
                argcount = len(argspec.args)

    if argcount is None:
        # Assume all args are handleable.
        argcount = maxargs

    return argcount


def iteriteratee(obj, iteratee=None, reverse=False):
    """Return iterative iteratee based on collection type."""
    if iteratee is None:
        cbk = pyd.identity
        argcount = 1
    else:
        cbk = pyd.iteratee(iteratee)
        argcount = getargcount(cbk, maxargs=3)

    items = iterator(obj)

    if reverse:
        items = reversed(tuple(items))

    for key, item in items:
        yield callit(cbk, item, key, obj, argcount=argcount), item, key, obj


def iterator(obj):
    """Return iterative based on object type."""
    if isinstance(obj, Mapping):
        return obj.items()
    elif hasattr(obj, "iteritems"):
        return obj.iteritems()  # noqa: B301
    elif hasattr(obj, "items"):
        return iter(obj.items())
    elif isinstance(obj, Iterable):
        return enumerate(obj)
    else:
        return getattr(obj, "__dict__", {}).items()


def base_get(obj, key, default=UNSET):
    """
    Safely get an item by `key` from a sequence or mapping object when `default` provided.

    Args:
        obj: Sequence or mapping to retrieve item from.
        key: Key or index identifying which item to retrieve.
        default: Default value to return if `key` not found in `obj`.

    Returns:
        `obj[key]`, `obj.key`, or `default`.

    Raises:
        KeyError: If `obj` is missing key, index, or attribute and no default value provided.
    """
    if isinstance(obj, dict):
        value = _base_get_dict(obj, key, default=default)
    elif not isinstance(obj, (Mapping, Sequence)) or (
        isinstance(obj, tuple) and hasattr(obj, "_fields")
    ):
        # Don't use getattr for dict/list objects since we don't want class methods/attributes
        # returned for them but do allow getattr for namedtuple.
        value = _base_get_object(obj, key, default=default)
    else:
        value = _base_get_item(obj, key, default=default)

    if value is UNSET:
        # Raise if there's no default provided.
        raise KeyError(f'Object "{repr(obj)}" does not have key "{key}"')

    return value


def _base_get_dict(obj, key, default=UNSET):
    value = obj.get(key, UNSET)
    if value is UNSET:
        value = default
        if not isinstance(key, int):
            # Try integer key fallback.
            try:
                value = obj.get(int(key), default)
            except Exception:
                pass
    return value


def _base_get_item(obj, key, default=UNSET):
    try:
        return obj[key]
    except Exception:
        pass

    if not isinstance(key, int):
        try:
            return obj[int(key)]
        except Exception:
            pass

    return default


def _base_get_object(obj, key, default=UNSET):
    value = _base_get_item(obj, key, default=UNSET)
    if value is UNSET:
        _raise_if_restricted_key(key)
        value = default
        try:
            value = getattr(obj, key)
        except Exception:
            pass
    return value


def _raise_if_restricted_key(key):
    # Prevent access to restricted keys for security reasons.
    if key in RESTRICTED_KEYS:
        raise KeyError(f"access to restricted key {key!r} is not allowed")


def base_set(obj, key, value, allow_override=True):
    """
    Set an object's `key` to `value`. If `obj` is a ``list`` and the `key` is the next available
    index position, append to list; otherwise, pad the list of ``None`` and then append to the list.

    Args:
        obj: Object to assign value to.
        key: Key or index to assign to.
        value: Value to assign.
        allow_override: Whether to allow overriding a previously set key.
    """
    if isinstance(obj, dict):
        if allow_override or key not in obj:
            obj[key] = value
    elif isinstance(obj, list):
        key = int(key)

        if key < len(obj):
            if allow_override:
                obj[key] = value
        else:
            if key > len(obj):
                # Pad list object with None values up to the index key, so we can append the value
                # into the key index.
                obj[:] = (obj + [None] * key)[:key]
            obj.append(value)
    elif (allow_override or not hasattr(obj, key)) and obj is not None:
        _raise_if_restricted_key(key)
        setattr(obj, key, value)

    return obj


def cmp(a, b):  # pragma: no cover
    """
    Replacement for built-in function ``cmp`` that was removed in Python 3.

    Note: Mainly used for comparison during sorting.
    """
    if a is None and b is None:
        return 0
    elif a is None:
        return -1
    elif b is None:
        return 1
    return (a > b) - (a < b)


def parse_iteratee(iteratee_keyword, *args, **kwargs):
    """Try to find iteratee function passed in either as a keyword argument or as the last
    positional argument in `args`."""
    iteratee = kwargs.get(iteratee_keyword)
    last_arg = args[-1]

    if iteratee is None and (
        callable(last_arg) or isinstance(last_arg, (dict, str)) or last_arg is None
    ):
        iteratee = last_arg
        args = args[:-1]

    return iteratee, args


class iterator_with_default(object):
    """A wrapper around an iterator object that provides a default."""

    def __init__(self, collection, default):
        self.iter = iter(collection)
        self.default = default

    def __iter__(self):
        return self

    def next_default(self):
        ret = self.default
        self.default = UNSET
        return ret

    def __next__(self):
        ret = next(self.iter, self.next_default())
        if ret is UNSET:
            raise StopIteration
        return ret

    next = __next__


def deprecated(func):  # pragma: no cover
    """
    This is a decorator which can be used to mark functions as deprecated.

    It will result in a warning being emitted when the function is used.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        warnings.warn(
            f"Call to deprecated function {func.__name__}.",
            category=DeprecationWarning,
            stacklevel=3,
        )
        return func(*args, **kwargs)

    return wrapper



import argparse
import ast
from collections import defaultdict, deque
from pathlib import Path
import typing as t


WRAPPER_KW = "RES"
INIT_FILE = "src/pydash/__init__.py"
BASE_MODULE = '''
# mypy: disable-error-code=misc
"""Generated from the `scripts/chaining_type_generator.py` script."""

import re
import typing as t
from typing_extensions import Concatenate, Literal, ParamSpec, Type

import pydash as pyd
from pydash.chaining.chaining import Chain
from pydash.types import *
from pydash.helpers import Unset, UNSET
from pydash.functions import (
    After,
    Ary,
    Before,
    CurryOne,
    CurryTwo,
    CurryThree,
    CurryFour,
    CurryFive,
    CurryRightOne,
    CurryRightTwo,
    CurryRightThree,
    CurryRightFour,
    CurryRightFive,
    Debounce,
    Disjoin,
    Flow,
    Iterated,
    Juxtapose,
    Negate,
    Once,
    Partial,
    Rearg,
    Spread,
    Throttle,
)
from pydash.utilities import MemoizedFunc

from _typeshed import (
    SupportsDunderGE,
    SupportsDunderGT,
    SupportsDunderLE,
    SupportsDunderLT,
    SupportsRichComparison,
    SupportsAdd,
    SupportsRichComparisonT,
    SupportsSub,
)

ValueT_co = t.TypeVar("ValueT_co", covariant=True)
T = t.TypeVar("T")
T1 = t.TypeVar("T1")
T2 = t.TypeVar("T2")
T3 = t.TypeVar("T3")
T4 = t.TypeVar("T4")
T5 = t.TypeVar("T5")
NumT = t.TypeVar("NumT", int, float, "Decimal")
NumT2 = t.TypeVar("NumT2", int, float, "Decimal")
NumT3 = t.TypeVar("NumT3", int, float, "Decimal")
CallableT = t.TypeVar("CallableT", bound=t.Callable[..., t.Any])
SequenceT = t.TypeVar("SequenceT", bound=t.Sequence[t.Any])
MutableSequenceT = t.TypeVar("MutableSequenceT", bound=t.MutableSequence[t.Any])
P = ParamSpec("P")


class {class_name}:
'''

FUNCTIONS_TO_SKIP = [
    # this is already a method of `Chain`
    "to_string",
]


def build_header(class_name: str) -> str:
    return BASE_MODULE.format(class_name=class_name)


def modules_and_api_funcs() -> t.Dict[str, t.List[str]]:
    """This is mostly so we don't have to import `pydash`"""

    with open(INIT_FILE, "r", encoding="utf-8") as source:
        tree = ast.parse(source.read())

    module_to_funcs = defaultdict(list)

    for node in ast.walk(tree):
        # TODO: maybe handle `Import` as well, not necessary for now
        if isinstance(node, ast.ImportFrom):
            for name in node.names:
                module_to_funcs[node.module].append(name.asname or name.name)

    return module_to_funcs


def is_overload(node: ast.FunctionDef) -> bool:
    return any(
        (
            (isinstance(decorator, ast.Name) and decorator.id == "overload")
            or (isinstance(decorator, ast.Attribute) and decorator.attr == "overload")
        )
        for decorator in node.decorator_list
    )


def returns_typeguard(node: ast.FunctionDef) -> bool:
    def is_constant_typeguard(cst: ast.expr) -> bool:
        return isinstance(cst, ast.Constant) and cst.value is not None and "TypeGuard" in cst.value

    def is_subscript_typeguard(sub: ast.expr) -> bool:
        return (
            isinstance(sub, ast.Subscript)
            and isinstance(sub.value, ast.Name)
            and "TypeGuard" in sub.value.id
        )

    return node.returns is not None and (
        is_constant_typeguard(node.returns) or is_subscript_typeguard(node.returns)
    )


def has_single_default_arg(node: ast.FunctionDef) -> bool:
    return len(node.args.args) == 1 and len(node.args.defaults) >= 1


def chainwrapper_args(
    node: ast.FunctionDef,
) -> t.Tuple[t.List[ast.expr], t.List[ast.keyword]]:
    # TODO: handle posonlyargs
    args: t.List[ast.expr] = [ast.Name(id=arg.arg) for arg in node.args.args[1:]]
    kwargs: t.List[ast.keyword] = [
        ast.keyword(arg=kw.arg, value=ast.Name(id=kw.arg)) for kw in node.args.kwonlyargs
    ]

    if node.args.vararg:
        args.append(ast.Starred(value=ast.Name(id=node.args.vararg.arg)))

    if node.args.kwarg:
        kwargs.append(ast.keyword(value=ast.Name(id=node.args.kwarg.arg)))

    return args, kwargs


def wrap_type(wrapper: ast.Subscript, to_wrap: ast.expr) -> ast.expr:
    if isinstance(wrapper.slice, ast.Tuple):
        slice = ast.Tuple(
            elts=[
                s if not (isinstance(s, ast.Name) and s.id == WRAPPER_KW) else to_wrap
                for s in wrapper.slice.elts
            ]
        )
    else:
        slice = to_wrap

    return ast.Subscript(
        value=wrapper.value,
        slice=slice,
    )


def get_first_arg(node: ast.FunctionDef) -> ast.arg:
    if node.args.args:
        return node.args.args[0]

    if node.args.vararg:
        return ast.arg(arg=node.args.vararg.arg, annotation=node.args.vararg.annotation)

    raise RuntimeError("Node should have a first argument")


def transform_function(node: ast.FunctionDef, wrapper: ast.Subscript) -> ast.FunctionDef:
    first_arg = get_first_arg(node)
    cw_args, cw_kwargs = chainwrapper_args(node)

    # case where we only have a vararg argument
    if not node.args.args:
        node.args.args.append(first_arg)

    if first_arg.annotation:
        first_arg.annotation = ast.Constant(
            value=ast.unparse(wrap_type(wrapper, first_arg.annotation))
        )

    first_arg.arg = "self"

    # we need to remove the first default arg as it is now the self argument
    if len(node.args.args) == len(node.args.defaults):
        node.args.defaults = node.args.defaults[1:]

    if node.returns:
        # TODO: `(some_arg: T) -> TypeGuard[T]` to `(some_arg: Any) -> bool`
        # TODO: otherwise we would get a `T` alone

        # change typeguard to bool as it is useless in a chain
        if returns_typeguard(node):
            node.returns = ast.Name(id="bool")

        node.returns = ast.Constant(value=ast.unparse(wrap_type(wrapper, node.returns)))

    if not is_overload(node):
        node.body = [
            ast.Return(
                value=ast.Call(
                    func=ast.Call(
                        func=ast.Name(id="self._wrap"),
                        args=[ast.Name(id=f"pyd.{node.name}")],
                        keywords=[],
                    ),
                    args=cw_args,
                    keywords=cw_kwargs,
                )
            )
        ]

    return node


def filename_from_module(module: str) -> str:
    return "src/pydash/chaining/chaining.py" if module == "chaining" else f"src/pydash/{module}.py"


def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--class_name",
        help="Name of the output class to put typed methods in",
        required=True,
    )
    parser.add_argument(
        "--output",
        type=Path,
        help="Path to the file to write the typed class to (probably a `.pyi` file)",
        required=True,
    )
    parser.add_argument(
        "--wrapper",
        help="The main generic class (eg. `Chain`)",
        required=True,
    )
    args = parser.parse_args()

    wrapper = args.wrapper + f"[{WRAPPER_KW}]"
    wrapper = ast.parse(wrapper).body[0]
    assert isinstance(wrapper, ast.Expr), "`wrapper` value should contain one expression"
    wrapper = wrapper.value
    assert isinstance(
        wrapper, ast.Subscript
    ), "`wrapper` value should contain one with one subscript"

    to_file = open(args.output, "w")
    to_file.write(build_header(args.class_name))

    module_to_funcs = modules_and_api_funcs()

    for module in module_to_funcs.keys():
        filename = filename_from_module(module)

        with open(filename, encoding="utf-8") as source:
            tree = ast.parse(source.read(), filename=filename)

        class_methods = deque()

        for node in ast.walk(tree):
            if isinstance(node, ast.ClassDef):
                class_methods.extend(f for f in node.body if isinstance(f, ast.FunctionDef))

            # skipping class methods
            if node in class_methods:
                class_methods.popleft()
                continue

            if (
                isinstance(node, ast.FunctionDef)
                and node.name in module_to_funcs[module]
                and (node.args.args or node.args.vararg)  # skipping funcs without args for now
                and not has_single_default_arg(node)  # skipping 1 default arg funcs
                and node.name not in FUNCTIONS_TO_SKIP
            ):
                new_node = transform_function(node, wrapper)
                to_file.write(" " * 4)
                to_file.write(ast.unparse(new_node).replace("\n", f"\n{' ' * 4}"))
                to_file.write("\n\n")
                if new_node.name.endswith("_") and not is_overload(new_node):
                    to_file.write(f"{' ' * 4}{new_node.name.rstrip('_')} = {new_node.name}")
                    to_file.write("\n\n")

    to_file.close()
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



import typing as t

import pytest

import pydash as _


class MyClass:
    def __init__(self) -> None:
        self.x = 5
        self.lst: t.List[int] = []

    def get_x(self) -> int:
        return self.x


@pytest.mark.mypy_testing
def test_mypy_assign() -> None:
    obj = {b"d": 5.5}
    reveal_type(_.assign(obj, {"a": 1}, {"b": 2}, {"c": 3}))  # R: builtins.dict[Union[builtins.bytes, builtins.str], Union[builtins.float, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_assign_with() -> None:
    def customizer(x: t.Union[int, None], y: int) -> float:
        return float(y) if x is None else float(x + y)

    reveal_type(_.assign_with({"a": 1}, {"b": 2}, {"a": 3}, customizer=customizer))  # R: builtins.dict[builtins.str, Union[builtins.int, builtins.float]]


@pytest.mark.mypy_testing
def test_mypy_callables() -> None:
    reveal_type(_.callables({"a": 1, "b": lambda: 2, "c": lambda: 3}))  # R: builtins.list[builtins.str]


@pytest.mark.mypy_testing
def test_mypy_clone() -> None:
    reveal_type(_.clone({"hello": 5}))  # R: builtins.dict[builtins.str, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_clone_with() -> None:
    x: t.Dict[str, t.Union[int, t.Dict[str, int]]] = {"a": 1, "b": 2, "c": {"d": 3}}

    def cbk(v: t.Union[int, t.Dict[str, int]], k: t.Union[str, None]) -> t.Union[int, None]:
        return v + 2 if isinstance(v, int) and k else None

    reveal_type(_.clone_with(x, cbk))  # R: builtins.dict[builtins.str, Union[builtins.int, builtins.dict[builtins.str, builtins.int], None]]


@pytest.mark.mypy_testing
def test_mypy_clone_deep() -> None:
    x: t.Dict[str, t.Union[int, t.Dict[str, int]]] = {"a": 1, "b": 2, "c": {"d": 3}}
    reveal_type(_.clone_deep(x))  # R: builtins.dict[builtins.str, Union[builtins.int, builtins.dict[builtins.str, builtins.int]]]


@pytest.mark.mypy_testing
def test_mypy_clone_deep_with() -> None:
    x: t.Dict[str, t.Union[int, t.Dict[str, int]]] = {"a": 1, "b": 2, "c": {"d": 3}}

    def cbk(v: t.Union[int, t.Dict[str, int]], k: t.Union[str, None]) -> t.Union[int, None]:
        return v + 2 if isinstance(v, int) and k else None

    reveal_type(_.clone_deep_with(x, cbk))  # R: builtins.dict[builtins.str, Union[builtins.int, builtins.dict[builtins.str, builtins.int], None]]


@pytest.mark.mypy_testing
def test_mypy_defaults() -> None:
    obj = {"a": 1}
    reveal_type(_.defaults(obj, {"b": 2}, {"c": 3}, {"a": 4}))  # R: builtins.dict[builtins.str, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_defaults_deep() -> None:
    obj = {"a": {"b": 1}}
    reveal_type(_.defaults_deep(obj, {"a": {"b": 2, "c": 3}}))  # R: builtins.dict[builtins.str, builtins.dict[builtins.str, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_find_key() -> None:
    def is_one(x: int) -> bool:
        return x == 1

    reveal_type(_.find_key({"a": 1, "b": 2, "c": 3}, is_one))  # R: Union[builtins.str, None]
    reveal_type(_.find_key([1, 2, 3, 4], is_one))  # R: Union[builtins.int, None]


@pytest.mark.mypy_testing
def test_mypy_find_last_key() -> None:
    def is_one(x: int) -> bool:
        return x == 1

    reveal_type(_.find_last_key({"a": 1, "b": 2, "c": 3}, is_one))  # R: Union[builtins.str, None]
    reveal_type(_.find_last_key([1, 2, 3, 1], is_one))  # R: Union[builtins.int, None]


@pytest.mark.mypy_testing
def test_mypy_for_in() -> None:
    def cb(v: int, k: str) -> None:
        return None

    reveal_type(_.for_in({"a": 1, "b": 2, "c": 3}, cb))  # R: builtins.dict[builtins.str, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_for_in_right() -> None:
    def cb(v: int) -> None:
        return None

    reveal_type(_.for_in_right([1, 2, 3, 4], cb))  # R: builtins.list[builtins.int]


@pytest.mark.mypy_testing
def test_mypy_get() -> None:
    reveal_type(_.get({}, "a.b.c"))  # R: Any
    reveal_type(_.get({"a": {"b": {"c": [1, 2, 3, 4]}}}, "a.b.c[1]"))  # R: Any
    reveal_type(_.get({"a": {"b": [0, {"c": [1, 2]}]}}, "a.b.1.c.2"))  # R: Any
    reveal_type(_.get(["a", "b"], 0))  # R: Union[builtins.str, None]
    reveal_type(_.get(["a", "b"], 0, "c"))  # R: builtins.str
    reveal_type(_.get(MyClass(), "x"))  # R: Any


@pytest.mark.mypy_testing
def test_mypy_has() -> None:
    reveal_type(_.has([1, 2, 3], 1))  # R: builtins.bool
    reveal_type(_.has({"a": 1, "b": 2}, "b"))  # R: builtins.bool
    reveal_type(_.has(MyClass(), "get_x"))  # R: builtins.bool


@pytest.mark.mypy_testing
def test_mypy_invert() -> None:
    reveal_type(_.invert({"a": 1, "b": 2, "c": 3}))  # R: builtins.dict[builtins.int, builtins.str]


@pytest.mark.mypy_testing
def test_mypy_invert_by() -> None:
    obj = {"a": 1, "b": 2, "c": 1}
    reveal_type(_.invert_by(obj))  # R: builtins.dict[builtins.int, builtins.list[builtins.str]]

    def group_prefix(x: int) -> str:
        return "group" + str(x)

    reveal_type(_.invert_by(obj, group_prefix))  # R: builtins.dict[builtins.str, builtins.list[builtins.str]]


@pytest.mark.mypy_testing
def test_mypy_invoke() -> None:
    obj = {"a": [{"b": {"c": [1, 2, 3, 4]}}]}
    reveal_type(_.invoke(obj, "a[0].b.c.pop", 1))  # R: Any
    reveal_type(_.invoke(MyClass(), "get_x"))  # R: Any


@pytest.mark.mypy_testing
def test_mypy_keys() -> None:
    reveal_type(_.keys([1, 2, 3]))  # R: builtins.list[builtins.int]
    reveal_type(_.keys({"a": 1, "b": 2}))  # R: builtins.list[builtins.str]
    reveal_type(_.keys(MyClass()))  # R: builtins.list[Any]


@pytest.mark.mypy_testing
def test_mypy_map_keys() -> None:
    def callback(value: int, key: str) -> str:
        return key * 2

    reveal_type(_.map_keys({"a": 1, "b": 2, "c": 3}, callback))  # R: builtins.dict[builtins.str, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_map_values() -> None:
    def times_two(x: int) -> str:
        return str(x * 2)

    reveal_type(_.map_values({"a": 1, "b": 2, "c": 3}, times_two))  # R: builtins.dict[builtins.str, builtins.str]
    reveal_type(_.map_values({"a": 1, "b": {"d": 4}, "c": 3}, {"d": 4}))  # R: builtins.dict[Any, Any]


@pytest.mark.mypy_testing
def test_mypy_map_values_deep() -> None:
    x = {"a": 1, "b": {"c": 2}}

    def times_two(x: int) -> int:
        return x * 2

    reveal_type(_.map_values_deep(x, times_two))  # R: Any


@pytest.mark.mypy_testing
def test_mypy_merge() -> None:
    obj = {"a": 2}
    reveal_type(_.merge(obj, {"a": 1}, {"b": 2, "c": 3}, {"d": 4}))  # R: builtins.dict[builtins.str, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_merge_with() -> None:
    cbk = lambda obj_val, src_val: obj_val + src_val
    obj1 = {"a": [1], "b": [2]}
    obj2 = {"a": [3], "b": [4]}
    reveal_type(_.merge_with(obj1, obj2, cbk))  # R: Any


@pytest.mark.mypy_testing
def test_mypy_omit() -> None:
    reveal_type(_.omit({"a": 1, "b": 2, "c": 3}, "b", "c"))  # R: builtins.dict[builtins.str, builtins.int]
    reveal_type(_.omit({"a": 1, "b": 2, "c": 3}, ["a", "c"]))  # R: builtins.dict[builtins.str, builtins.int]
    reveal_type(_.omit([1, 2, 3, 4], 0, 3))  # R: builtins.dict[builtins.int, builtins.int]
    reveal_type(_.omit({"a": {"b": {"c": "d"}}}, "a.b.c"))  # R: builtins.dict[builtins.str, builtins.dict[builtins.str, builtins.dict[builtins.str, builtins.str]]]
    reveal_type(_.omit(MyClass(), "x"))  # R: builtins.dict[Any, Any]


@pytest.mark.mypy_testing
def test_mypy_omit_by() -> None:
    def is_int(v: t.Union[str, int]) -> bool:
        return isinstance(v, int)

    obj: t.Dict[str, t.Union[str, int]] = {"a": 1, "b": "2", "c": 3}
    reveal_type(_.omit_by(obj, is_int))  # R: builtins.dict[builtins.str, Union[builtins.str, builtins.int]]
    reveal_type(_.omit_by([1, 2, 3, 4], is_int))  # R: builtins.dict[builtins.int, builtins.int]
    reveal_type(_.omit_by(MyClass(), is_int))  # R: builtins.dict[Any, Any]


@pytest.mark.mypy_testing
def test_mypy_parse_int() -> None:
    reveal_type(_.parse_int("5"))  # R: Union[builtins.int, None]
    reveal_type(_.parse_int("12", 8))  # R: Union[builtins.int, None]


@pytest.mark.mypy_testing
def test_mypy_pick() -> None:
    reveal_type(_.pick({"a": 1, "b": 2, "c": 3}, "a", "b"))  # R: builtins.dict[builtins.str, builtins.int]
    reveal_type(_.pick(MyClass(), "x"))  # R: builtins.dict[Any, Any]


@pytest.mark.mypy_testing
def test_mypy_pick_by() -> None:
    def is_int(v: t.Union[int, str]) -> bool:
        return isinstance(v, int)

    obj: t.Dict[str, t.Union[int, str]] = {"a": 1, "b": "2", "c": 3}

    reveal_type(_.pick_by(obj, is_int))  # R: builtins.dict[builtins.str, Union[builtins.int, builtins.str]]
    reveal_type(_.pick(MyClass(), lambda v: isinstance(v, int)))  # R: builtins.dict[Any, Any]


@pytest.mark.mypy_testing
def test_mypy_rename_keys() -> None:
    reveal_type(_.rename_keys({"a": 1, "b": 2, "c": 3}, {"a": "A", "b": "B"}))  # R: builtins.dict[builtins.str, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_set_() -> None:
    reveal_type(_.set_({}, "a.b.c", 1))  # R: builtins.dict[Never, Never]
    reveal_type(_.set_(MyClass(), "x", 10))  # R: tests.pytest_mypy_testing.test_objects.MyClass


@pytest.mark.mypy_testing
def test_mypy_set_with() -> None:
    reveal_type(_.set_with({}, "[0][1]", "a", lambda: {}))  # R: builtins.dict[Never, Never]
    reveal_type(_.set_with(MyClass(), "x", lambda: 10))  # R: tests.pytest_mypy_testing.test_objects.MyClass


@pytest.mark.mypy_testing
def test_mypy_to_boolean() -> None:
    reveal_type(_.to_boolean("true"))  # R: Union[builtins.bool, None]


@pytest.mark.mypy_testing
def test_mypy_to_dict() -> None:
    obj = {"a": 1, "b": 2}
    reveal_type(_.to_dict(obj))  # R: builtins.dict[builtins.str, builtins.int]
    reveal_type(_.to_dict(MyClass()))  # R: builtins.dict[Any, Any]
    reveal_type(_.to_dict([1, 2, 3, 4]))  # R: builtins.dict[builtins.int, builtins.int]
    reveal_type(_.to_dict([(1, 2), (3, 4)]))  # R: builtins.dict[builtins.int, Tuple[builtins.int, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_to_integer() -> None:
    reveal_type(_.to_integer(3.2))  # R: builtins.int
    reveal_type(_.to_integer("3.2"))  # R: builtins.int
    reveal_type(_.to_integer("invalid"))  # R: builtins.int


@pytest.mark.mypy_testing
def test_mypy_to_list() -> None:
    reveal_type(_.to_list({"a": 1, "b": 2, "c": 3}))  # R: builtins.list[builtins.int]
    reveal_type(_.to_list((1, 2, 3, 4)))  # R: builtins.list[builtins.int]
    reveal_type(_.to_list(1))  # R: builtins.list[builtins.int]
    reveal_type(_.to_list([1]))  # R: builtins.list[builtins.int]
    reveal_type(_.to_list(a for a in [1, 2, 3]))  # R: builtins.list[builtins.int]
    reveal_type(_.to_list("cat"))  # R: builtins.list[builtins.str]
    reveal_type(_.to_list("cat", split_strings=False))  # R: builtins.list[builtins.str]


@pytest.mark.mypy_testing
def test_mypy_to_number() -> None:
    reveal_type(_.to_number("1234.5678"))  # R: Union[builtins.float, None]
    reveal_type(_.to_number("1234.5678", 4))  # R: Union[builtins.float, None]
    reveal_type(_.to_number(1, 2))  # R: Union[builtins.float, None]


@pytest.mark.mypy_testing
def test_mypy_to_pairs() -> None:
    reveal_type(_.to_pairs([1, 2, 3, 4]))  # R: builtins.list[Tuple[builtins.int, builtins.int]]
    reveal_type(_.to_pairs({"a": 1}))  # R: builtins.list[Tuple[builtins.str, builtins.int]]
    reveal_type(_.to_pairs(MyClass()))  # R: builtins.list[Any]


@pytest.mark.mypy_testing
def test_mypy_to_string() -> None:
    reveal_type(_.to_string(1))  # R: builtins.str
    reveal_type(_.to_string(None))  # R: builtins.str
    reveal_type(_.to_string([1, 2, 3]))  # R: builtins.str
    reveal_type(_.to_string("a"))  # R: builtins.str


@pytest.mark.mypy_testing
def test_mypy_transform() -> None:
    def build_list(acc: t.List[t.Tuple[int, int]], v: int, k: int) -> None:
        return acc.append((k, v))

    base_list: t.List[t.Tuple[int, int]] = []
    reveal_type(_.transform([1, 2, 3, 4], build_list, base_list))  # R: builtins.list[Tuple[builtins.int, builtins.int]]


@pytest.mark.mypy_testing
def test_mypy_update() -> None:
    reveal_type(_.update({}, ["a", "b"], lambda value: value))  # R: builtins.dict[Any, Any]
    reveal_type(_.update([], [0, 0], lambda value: 1))  # R: builtins.list[Any]
    reveal_type(_.update(MyClass(), "x", lambda value: 10))  # R: tests.pytest_mypy_testing.test_objects.MyClass


@pytest.mark.mypy_testing
def test_mypy_update_with() -> None:
    reveal_type(_.update_with({}, "[0][1]", lambda x: "a", lambda x: {}))  # R: builtins.dict[Any, Any]
    reveal_type(_.update_with([], [0, 0], lambda x: 1, lambda x: []))  # R: builtins.list[Any]
    reveal_type(_.update_with(MyClass(), "lst.0", lambda value: 10, lambda x: []))  # R: tests.pytest_mypy_testing.test_objects.MyClass


@pytest.mark.mypy_testing
def test_mypy_unset() -> None:
    reveal_type(_.unset({"a": [{"b": {"c": 7}}]}, "a[0].b.c"))  # R: builtins.bool


@pytest.mark.mypy_testing
def test_mypy_values() -> None:
    reveal_type(_.values({"a": "a", "b": "b", "c": "c"}))  # R: builtins.list[builtins.str]
    reveal_type(_.values({"a": 1, "b": 2, "c": 3}))  # R: builtins.list[builtins.int]
    reveal_type(_.values([2, 4, 6, 8]))  # R: builtins.list[builtins.int]
    reveal_type(_.values(MyClass()))  # R: builtins.list[Any]


@pytest.mark.mypy_testing
def test_mypy_apply() -> None:
    reveal_type(_.apply("1", lambda x: int(x)))  # R: builtins.int
    reveal_type(_.apply(1, lambda x: x + 1))  # R: builtins.int
    reveal_type(_.apply("hello", lambda x: x.upper()))  # R: builtins.str


@pytest.mark.mypy_testing
def test_mypy_apply_if() -> None:
    reveal_type(_.apply_if("5", lambda x: int(x), lambda x: x.isdecimal()))  # R: Union[builtins.str, builtins.int]


@pytest.mark.mypy_testing
def test_mypy_apply_if_not_none() -> None:
    reveal_type(_.apply_if_not_none(1, lambda x: x + 1))  # R: Union[builtins.int, None]
    reveal_type(_.apply_if_not_none(None, lambda x: x + 1))  # R: Union[builtins.int, None]
    reveal_type(_.apply_if_not_none("hello", lambda x: x.upper()))  # R: Union[builtins.str, None]


@pytest.mark.mypy_testing
def test_mypy_apply_catch() -> None:
    reveal_type(_.apply_catch(5, lambda x: x / 0, [ZeroDivisionError]))  # R: Union[builtins.int, builtins.float]
    reveal_type(_.apply_catch(5, lambda x: x / 0, [ZeroDivisionError], "error"))  # R: Union[builtins.float, builtins.str]



import typing as t

import pydash as _


def typed_function(row: int, index: int, col: t.List[t.Any]):
    return row + 1


def test_annotated_iteratee():
    assert _.map_([1, 2], typed_function) == [2, 3]



from pathlib import Path
import re
import sys
import typing as t


class DocString(t.NamedTuple):
    content: str
    function_name: str

    def example_block(self) -> t.Union[t.Iterable[str], None]:
        expression_re = re.compile(r"Examples?:\n+((?:\s{8}.+?\n)+)")

        match = expression_re.search(self.content)
        if not match:
            return None

        example = match.group(1)

        return (line.strip() for line in example.strip().splitlines())


def docstrings(path: Path) -> t.Iterable[DocString]:
    docstring_re = re.compile(r"def (.+?)\((?:.|\n)+?:[\s\n]+?r?\"\"\"((?:.|\s|\n)+?)\"\"\"")

    with open(path) as f:
        text = f.read()
        docstrings = docstring_re.finditer(text)

    return map(
        lambda match: DocString(function_name=match.group(1), content=match.group(2)), docstrings
    )


def generate_test_function(docstring: DocString) -> str:
    if not (example_block := docstring.example_block()):
        return ""

    built_function = ""

    built_function += "@pytest.mark.mypy_testing\n"
    built_function += f"def test_mypy_{docstring.function_name}() -> None:\n"

    for line in map(lambda line_: line_.strip(), example_block):
        if not line:
            continue

        to_be_revealed = line.replace(">>> ", "")
        if line.startswith(">>> "):
            to_be_revealed = f"_.{to_be_revealed}"

        built_function += f"    reveal_type({to_be_revealed})  # R:\n"

    return built_function


def main(path: Path) -> str:
    imports = "import pytest\n\n"
    imports += "import pydash as _\n\n\n"

    return imports + "\n\n".join(map(generate_test_function, docstrings(path)))


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("filename", help="path to python file", type=Path)
    parser.add_argument("output", help="path to output file", type=Path)
    args = parser.parse_args()

    if not args.filename.exists():
        print(f"`{args.filename}` does not exist")
        sys.exit(1)

    args.output.write_text(main(args.filename))
